# package
#coding=utf-8
import chardet,re
import geventcurl as pycurl
from cStringIO import StringIO

class Curl:
    def __init__(self,url,**kw):
        self.url = self._get_yes_http(self._str_decode(url))
        self.args = kw 
        self._is_run = False
        self.err = None
        self._html = None
        self._code = None
        self._respip = None 
        self._headers = {}
        self._download_size = None
        self.c = None
        
    def __del__(self):
        if self.c:
            try:
                self.c.close()
            except:
                pass
            
            
    def _init(self): 
        self.body = StringIO()
        self.header = StringIO()
        c = pycurl.Curl() 
        c.setopt(pycurl.URL,self._str_decode(self.url))
        if self.url.lower().find('https://') == 0:
            c.setopt(pycurl.SSL_VERIFYHOST,False)
            c.setopt(pycurl.SSL_VERIFYPEER,False)
        c.setopt(pycurl.NOSIGNAL,1) 
        c.setopt(pycurl.NOPROGRESS, 1)
        c.setopt(pycurl.ENCODING,'gzip,deflate')
        agent = self.args.get('agent',None)
        if agent == 'no':pass
        else:
            if not agent: agent = "Mozilla/4.0 (compatible; MSIE 8.0; Windows NT 5.1; SV1; .NET CLR 1.1.4322)" 
            c.setopt(pycurl.USERAGENT, agent)
            
        '-----------------超时设置' 
        timeout = self.args.get('timeout',15)
        c.setopt(pycurl.TIMEOUT, timeout)    #超时时间
        c.setopt(pycurl.DNS_USE_GLOBAL_CACHE,1) 
        '-------------------'
        is_redirect = self.args.get('redirs',False)
        if is_redirect:
            c.setopt( pycurl.FOLLOWLOCATION, 1 )
            c.setopt( pycurl.MAXREDIRS, 5) 
        else:
            c.setopt( pycurl.FOLLOWLOCATION, 0)
            
        method = self.args.get('method','get').lower()
        #if method!='head':
            #c.setopt(pycurl.HEADER,0)
        c.setopt(pycurl.WRITEFUNCTION, self.body.write)
        '==================='
        if method == 'head': 
            c.setopt(pycurl.NOBODY,1)
        '无论如何都要将头部内容取回来'
        c.setopt(pycurl.HEADERFUNCTION, self.header.write)
        '================='
        #if method == 'get':
        #elif method == 'head': 
        #c.setopt(pycurl.NOBODY, 1)  
        if method == 'post':
            c.setopt(pycurl.POST,1)
            post_content = self.args.get('post_content',None)
            if post_content: c.setopt(pycurl.POSTFIELDS,post_content)
        else:
            c.setopt(pycurl.HTTPGET,1)
            
        cookie = self.args.get('cookie',None)
        if cookie:c.setopt(pycurl.COOKIE,cookie) 
        user = self.args.get('user',None)
        psw = self.args.get('psw',None)
        if user and psw:
            c.setopt(pycurl.HTTPAUTH,pycurl.HTTPAUTH_BASIC) 
            c.setopt(pycurl.USERPWD,user + ':' + psw) 
        
        self.c = c
        
    def _str_decode(self,val):
        '''
        将所有unicode转换成字符串
        '''
        if val and type(val) == unicode:
            val = val.encode('utf-8')
        return val
    
    def _get_yes_http(self,url):
        '''
        还原网址的http
        '''
        if url.find('://')==-1:
            url = 'http://' + url
        return url
    
    def run(self):
        '''
        一定要先调用这个才能有c
        ''' 
        try:
            self._init()
            self._is_run = True
            self.c.perform()
        except Exception,e:
            self.err = e
            return False
        return True
    @property
    def download_size(self):
        '''
        下载内容大小
        '''
        if self._download_size is None:self._download_size = self.c.getinfo(pycurl.SIZE_DOWNLOAD) 
        return self._download_size
    @property
    def headers(self):
        '响应头'
        if self._headers == {}:
            tmp = self.header.getvalue().split("\r\n")
            self._headers['ver'] = tmp[0]
            for t in tmp[1:]: 
                item = t.split(':')
                if len(item) == 2: 
                    self._headers[item[0].strip().lower()] = item[1].strip() 
        return self._headers
    @property
    def respip(self):
        '''
        服务器IP
        '''
        if self._respip is None:self._respip = str(self.c.getinfo(pycurl.PRIMARY_IP))
        return self._respip
    
    @property
    def code(self):
        '''
        响应码
        '''
        if self._code is None:self._code = self.c.getinfo(pycurl.HTTP_CODE)
        return self._code
    
    @property
    def html(self):
        '''
        取得正文
        '''
        if not self._is_run: raise Exception('请先运行RUN方法，再调用此属性')
        if self._html : return self._html
        self._html = self.body.getvalue()
        is_decode = self.args.get('decode',False)
        if is_decode:
            self._html = self._decode(self._html) 
        return self._html
    
    def _decode(self,content,plist = None):
        '''
        对正文进行相关编码
        '''
        def getcode(content,encoding):
            if encoding: 
                encoding = encoding.upper() 
            else:
                encoding = 'GB2312'
            if encoding in ['GB2312','GBK','ZH-CN', 'EUC-TW']: encoding = 'GB18030' 
             
            try:
                ret = content.decode(encoding,'ignore').encode('utf-8')  #str(self._content, encoding, errors='replace')
            except:
                return None
            
            return ret
         
        is_retry = False
        encoding = 'none'
        if plist and plist.find('=')!=-1: encoding = plist.split('=')[1].strip()

        if encoding.lower() == 'none': 
            p = re.compile( r'(<meta.*?[^>]*>)',re.IGNORECASE)
            vals = re.findall(p,content) 
            p_content = ''.join(vals) 
            content_ary = p_content.split('charset')
            if len(content_ary) > 1 : 
                content_ary = content_ary[1].replace('=','').strip()
                content_ary = content_ary.split('>')
                encoding = content_ary[0] 
                encoding = encoding.replace('"','').replace("'",'').replace('/','').strip() 
                is_retry = True
            else:
                encoding = chardet.detect(content)['encoding'] #response._detected_encoding()
                is_retry = False
        else: 
            is_retry = True
                
        if is_retry:
            if encoding.find(';')!=-1:
                content_ary = encoding.split(';') 
                encoding = content_ary[0]
            if encoding.find(' ')!=-1:
                content_ary = encoding.split(' ') 
                encoding = content_ary[0]
        '-----------------------------'
        ret = getcode(content,encoding)
        
        if is_retry and ret is None:
            encoding = chardet.detect(content)['encoding'] 
            ret = getcode(content,encoding)
        return ret
    
    def close(self): 
        if self._is_run: self.c.close()



#coding=utf-8
"""Implementation of pycurl's "easy" interface that uses pycurl's multi interface + gevent.
"""
# parts of code from Tornado's curl_httpclient.py
import logging
import gevent
assert gevent.version_info[:2] >= (0, 14), 'Gevent 0.14 or older required. Your version is %s' % gevent.__version__
from gevent.core import EVENTS, READ, WRITE
from gevent.hub import Waiter


pycurl = __import__('pycurl2')
for key, value in pycurl.__dict__.iteritems():
    if not key.startswith('_'):
        key = {'Curl': '_Curl'}.get(key, key)
        globals()[key] = value


class GeventCurl(object):
    """Integrate Curl's Multi interface into gevent's loop."""
    # QQQ cannot inherit from CurlMulti for some reason

    def __init__(self, loop=None):
        if loop is None:
            loop = gevent.get_hub().loop
        self.loop = loop
        self._watchers = {}
        self._timeout = None
        self._obj = CurlMulti()
        self.setopt(pycurl.M_TIMERFUNCTION, self._set_timeout)
        self.setopt(pycurl.M_SOCKETFUNCTION, self._set_socket)
        try:
            self._socket_action = self._obj.socket_action
        except AttributeError:
            # socket_action is found in pycurl since 7.18.2 (it's been
            # in libcurl longer than that but wasn't accessible to
            # python).
            logging.warning("socket_action method missing from pycurl; "
                            "falling back to socket_all. Upgrading "
                            "libcurl and pycurl will improve performance")
            self._socket_action = \
                lambda fd, action: self._obj.socket_all()

    def __getattr__(self, item):
        return getattr(self._obj, item)

    def _set_timeout(self, msecs):
        """Called by libcurl to schedule a timeout."""
        if self._timeout is not None:
            self._timeout.stop()
        self._timeout = self.loop.timer(msecs)
        self._timeout.start(self._handle_timeout)

    def add_handle(self, curl):
        curl = getattr(curl, '_obj', curl)
        self._obj.add_handle(curl)
        self._set_timeout(0)

    def remove_handle(self, curl):
        curl = getattr(curl, '_obj', curl)
        self._obj.remove_handle(curl)

    def _set_socket(self, event, fd, multi, data):
        """Called by libcurl when it wants to change the file descriptors it cares about."""
        # pycurl's event happen to match libev's event
        if event == pycurl.POLL_REMOVE or not event:
            watcher = self._watchers.pop(fd, None)
            if watcher is not None:
                watcher.stop()
        else:
            watcher = self._watchers.get(fd)
            if watcher is None:
                watcher = self.loop.io(fd, event)
                watcher.start(self._handle_events, EVENTS, fd)
                self._watchers[fd] = watcher
            else:
                if watcher.events != event:
                    watcher.stop()
                    watcher.events = event
                    watcher.start(self._handle_events, EVENTS, fd)

    def _handle_timeout(self):
        """Called by IOLoop when the requested timeout has passed."""
        if self._timeout is not None:
            self._timeout.stop()
            self._timeout = None
        while True:
            try:
                ret, num_handles = self._obj.socket_action(pycurl.SOCKET_TIMEOUT, 0)
            except pycurl.error, e:
                ret = e.args[0]
            if ret != pycurl.E_CALL_MULTI_PERFORM:
                break
        self._finish_pending_requests()

        # In theory, we shouldn't have to do this because curl will
        # call _set_timeout whenever the timeout changes. However,
        # sometimes after _handle_timeout we will need to reschedule
        # immediately even though nothing has changed from curl's
        # perspective. This is because when socket_action is
        # called with SOCKET_TIMEOUT, libcurl decides internally which
        # timeouts need to be processed by using a monotonic clock
        # (where available) while tornado uses python's time.time()
        # to decide when timeouts have occurred. When those clocks
        # disagree on elapsed time (as they will whenever there is an
        # NTP adjustment), tornado might call _handle_timeout before
        # libcurl is ready. After each timeout, resync the scheduled
        # timeout with libcurl's current state.
        new_timeout = self._obj.timeout()
        if new_timeout != -1:
            self._set_timeout(new_timeout)

    def _handle_events(self, events, fd):
        action = 0
        if events & READ:
            action |= pycurl.CSELECT_IN
        if events & WRITE:
            action |= pycurl.CSELECT_OUT
        while True:
            try:
                ret, num_handles = self._socket_action(fd, action)
            except pycurl.error, e:
                ret = e.args[0]
            if ret != pycurl.E_CALL_MULTI_PERFORM:
                break
        self._finish_pending_requests()

    def _finish_pending_requests(self):
        """Process any requests that were completed by the lastcall to multi.socket_action."""
        while True:
            num_q, ok_list, err_list = self._obj.info_read()
            for curl in ok_list:
                curl.waiter.switch()
            for curl, errnum, errmsg in err_list:
                curl.waiter.throw(Exception('%s %s' % (errnum, errmsg)))
            if num_q == 0:
                break


class cached_class_property(object):

    def __init__(self, func, name=None, doc=None):
        name = name or func.__name__
        self.__name__ = '_cached_class_property_' + name
        self.__module__ = func.__module__
        self.__doc__ = doc or func.__doc__
        self.func = func

    def __get__(self, _obj, cls, _missing=object()):
        value = getattr(cls, self.__name__, _missing)
        if value is _missing:
            value = self.func(cls)
            setattr(cls, self.__name__, value)
        return value


class Curl(object):

    _multi_class = GeventCurl

    @cached_class_property  # should be threadlocal_cached_class_property XXX
    def _multi(cls):
        return cls._multi_class()

    def __init__(self):
        self._obj = _Curl()

    def __getattr__(self, item):
        return getattr(self._obj, item)

    def perform(self):
        assert getattr(self._obj, 'waiter', None) is None, 'This curl object is already used by another greenlet'
        waiter = self._obj.waiter = Waiter()
        try:
            self._multi.add_handle(self)
            try:
                return waiter.get()
            finally:
                self._multi.remove_handle(self)
        finally:
            del self._obj.waiter


##############################################################################
#
# Copyright (c) 2013 Web Consulting and Contributors.
# All Rights Reserved.
#
##############################################################################

import os
import sys

from setuptools import setup, find_packages

py_version = sys.version_info[:2]


if py_version < (2, 6):
    raise RuntimeError('On Python 2, Gcrul  requires Python 2.6 or better')

here = os.path.abspath(os.path.dirname(__file__))
try:
    README = open(os.path.join(here, 'README.rst')).read()
    CHANGES = open(os.path.join(here, 'CHANGES.txt')).read()
except IOError:
    README = CHANGES = ''

install_requires=[
    'setuptools',
    'gevent'
    ]


setup(name='Gcurl',
      version='1.0',
      description=('The descusr web application development framework, a '
                   'gevent curl project'),
      long_description=README + '\n\n' +  CHANGES,
      classifiers=[
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: WSGI",
        "License :: Repoze Public License",
        ],
      keywords='gevent curl, Gcurl',
      author="descusr",
      author_email="descusr@163.com",
      url="http://www.piadu.com",
      license="BSD-derived (http://www.repoze.org/LICENSE.txt)",
      packages=find_packages(),
      include_package_data=True,
      zip_safe=False,
      install_requires = install_requires
      )

